﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2018 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Optimization {
  /// <summary>
  /// A base class for heuristic optimization algorithms.
  /// </summary>
  [Item("Heuristic Optimization Algorithm", "A base class for heuristic optimization algorithms.")]
  [StorableClass]
  public abstract class HeuristicOptimizationAlgorithm : Algorithm {
    public new IHeuristicOptimizationProblem Problem {
      get { return (IHeuristicOptimizationProblem)base.Problem; }
      set { base.Problem = value; }
    }

    [Storable]
    private TimeSpan executionTime;
    private DateTime startDate;
    public override TimeSpan ExecutionTime {
      get { return executionTime + (ExecutionState == ExecutionState.Started ? DateTime.UtcNow - startDate : TimeSpan.Zero); }
    }

    protected HeuristicOptimizationAlgorithm() : base() {
      executionTime = TimeSpan.Zero;
    }
    protected HeuristicOptimizationAlgorithm(string name) : base(name) {
      executionTime = TimeSpan.Zero;
    }
    protected HeuristicOptimizationAlgorithm(string name, ParameterCollection parameters) : base(name, parameters) {
      executionTime = TimeSpan.Zero;
    }
    protected HeuristicOptimizationAlgorithm(string name, string description) : base(name, description) {
      executionTime = TimeSpan.Zero;
    }
    protected HeuristicOptimizationAlgorithm(string name, string description, ParameterCollection parameters) : base(name, description, parameters) {
      executionTime = TimeSpan.Zero;
    }

    [StorableConstructor]
    protected HeuristicOptimizationAlgorithm(bool deserializing) : base(deserializing) { }
    protected HeuristicOptimizationAlgorithm(HeuristicOptimizationAlgorithm original, Cloner cloner)
      : base(original, cloner) {
      executionTime = original.executionTime;
    }

    public override void Prepare() {
      executionTime = TimeSpan.Zero;
      base.Prepare();
    }

    public override void Start(CancellationToken cancellationToken) {
      startDate = DateTime.UtcNow;
      base.Start(cancellationToken);
    }

    #region Events
    protected override void DeregisterProblemEvents() {
      Problem.SolutionCreatorChanged -= new EventHandler(Problem_SolutionCreatorChanged);
      Problem.EvaluatorChanged -= new EventHandler(Problem_EvaluatorChanged);
      base.DeregisterProblemEvents();
    }
    protected override void RegisterProblemEvents() {
      base.RegisterProblemEvents();
      Problem.SolutionCreatorChanged += new EventHandler(Problem_SolutionCreatorChanged);
      Problem.EvaluatorChanged += new EventHandler(Problem_EvaluatorChanged);
    }

    protected virtual void Problem_SolutionCreatorChanged(object sender, EventArgs e) { }
    protected virtual void Problem_EvaluatorChanged(object sender, EventArgs e) { }
    #endregion

    protected override void OnPaused() {
      var elapsed = DateTime.UtcNow - startDate;
      executionTime += elapsed;
      base.OnPaused();
    }

    protected override void OnStopped() {
      if (ExecutionState == ExecutionState.Started) {
        var elapsed = DateTime.UtcNow - startDate;
        executionTime += elapsed;
      }
      base.OnStopped();
    }
  }
}
